import asyncio
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from mcp.client.stdio import StdioServerParameters

from swarmx import (
    Agent,
    ToolRegistry,
    does_function_need_context,
)

pytestmark = pytest.mark.anyio


def test_does_function_need_context():
    # Test with function that needs context variables
    def func_with_ctx(context_variables: dict):
        return "Hello"

    assert does_function_need_context(func_with_ctx) is True

    # Test with function that doesn't need context variables
    def func_without_ctx(name: str):
        return f"Hello {name}"

    assert does_function_need_context(func_without_ctx) is False


@pytest.fixture
def tool_registry():
    registry = ToolRegistry()
    yield registry
    asyncio.run(registry.close())


def test_tool_registry_properties(tool_registry):
    # Test the tools property
    assert tool_registry.tools == {}

    # Add a function
    def test_func():
        return "Test"

    tool_registry.add_function(test_func)

    # Check that the function was added correctly
    assert len(tool_registry.tools) == 1
    assert tool_registry.tools["test_func"]["function"]["name"] == "test_func"


async def test_add_function_and_call_tool(tool_registry):
    # Define a test function
    def test_func(name: str):
        return f"Hello, {name}!"

    # Add the function to the registry
    tool_registry.add_function(test_func)

    # Call the tool
    result = await tool_registry.call_tool("test_func", {"name": "World"})

    # Check the result
    assert result == "Hello, World!"


async def test_call_tool_with_context(tool_registry):
    # Define a test function that needs context variables
    def test_func_with_ctx(name: str, context_variables: dict):
        return f"Hello, {name}! User ID: {context_variables.get('user_id')}"

    # Add the function to the registry
    tool_registry.add_function(test_func_with_ctx)

    # Call the tool with context variables
    result = await tool_registry.call_tool(
        "test_func_with_ctx", {"name": "World"}, {"user_id": "12345"}
    )

    # Check the result
    assert result == "Hello, World! User ID: 12345"


async def test_call_tool_async_function(tool_registry):
    # Define an async test function
    async def async_test_func(name: str):
        await asyncio.sleep(0.01)  # Simulate async operation
        return f"Hello async, {name}!"

    # Add the function to the registry
    tool_registry.add_function(async_test_func)

    # Call the tool
    result = await tool_registry.call_tool("async_test_func", {"name": "World"})

    # Check the result
    assert result == "Hello async, World!"


async def test_call_tool_agent_return(tool_registry):
    # Define a function that returns an Agent
    def agent_return_func():
        return Agent(name="TestAgent", instructions="Test instructions")

    # Add the function to the registry
    tool_registry.add_function(agent_return_func)

    # Call the tool
    result = await tool_registry.call_tool("agent_return_func", {})

    # Check the result
    assert isinstance(result, Agent)
    assert result.name == "TestAgent"


async def test_call_tool_not_found(tool_registry):
    # Attempt to call a tool that doesn't exist
    with pytest.raises(ValueError, match="Tool nonexistent_tool not found"):
        await tool_registry.call_tool("nonexistent_tool", {})


async def test_add_mcp_server():
    with (
        patch("swarmx.stdio_client") as mock_stdio_client,
        patch("swarmx.ClientSession") as mock_client_session,
    ):
        # Setup mocks
        mock_client = AsyncMock()
        mock_client.initialize = AsyncMock()
        mock_client.list_tools = AsyncMock()

        tool = MagicMock()
        tool.name = "test_tool"
        tool.model_dump.return_value = {
            "name": "test_tool",
            "inputSchema": {"type": "object", "properties": {}},
        }

        mock_client.list_tools.return_value.tools = [tool]
        mock_client_session.return_value.__aenter__.return_value = mock_client

        mock_read_stream = AsyncMock()
        mock_write_stream = AsyncMock()
        mock_stdio_client.return_value.__aenter__.return_value = (
            mock_read_stream,
            mock_write_stream,
        )

        # Create a registry and add an MCP server
        registry = ToolRegistry()
        await registry.add_mcp_server(
            "test_server", StdioServerParameters(command="echo", args=["test"])
        )

        # Verify that the client was initialized and tools were fetched
        mock_client.initialize.assert_called_once()
        mock_client.list_tools.assert_called_once()

        # Verify that the tool was added correctly
        assert ("test_server", "test_tool") in registry.mcp_tools
        assert registry._tools["test_tool"] == "test_server"

        # Cleanup
        await registry.close()



"""
This is an example of using SwarmX to generate ideas for research experiments.

Modified from https://github.com/SakanaAI/AI-Scientist
"""

from swarmx import Agent, Swarm

reasoning_models = ["deepseek-r1", "deepseek-reasoner", "claude-3.7-sonnet", "qwq"]

idea_generator = Agent(
    name="Idea Generator",
    instructions="""{{ task_description }}
<experiment.py>
{{ code }}
</experiment.py>

{% if prev_ideas %}
Here are the ideas that you have already generated:

'''
{% for idea in prev_ideas %}
{prev_ideas_string}
{% endfor %}
'''
{% endif %}

Come up with the next impactful and creative idea for research experiments and directions you can feasibly investigate with the code provided.
Note that you will not have access to any additional resources or datasets.
Make sure any idea is not overfit the specific training dataset or model, and has wider significance.

Respond in the following format:

{% if model not in reasoning_models %}
THOUGHT:
<THOUGHT>

NEW IDEA JSON:
{% endif %}
```json
<JSON>
```

{% if model not in reasoning_models %}
In <THOUGHT>, first briefly discuss your intuitions and motivations for the idea. Detail your high-level plan, necessary design choices and ideal outcomes of the experiments. Justify how the idea is different from the existing ones.
{% endif %}

In <JSON>, provide the new idea in JSON format with the following fields:
- "Name": A shortened descriptor of the idea. Lowercase, no spaces, underscores allowed.
- "Title": A title for the idea, will be used for the report writing.
- "Experiment": An outline of the implementation. E.g. which functions need to be added or modified, how results will be obtained, ...
- "Interestingness": A rating from 1 to 10 (lowest to highest).
- "Feasibility": A rating from 1 to 10 (lowest to highest).
- "Novelty": A rating from 1 to 10 (lowest to highest).

Be cautious and realistic on your ratings.
This JSON will be automatically parsed, so ensure the format is precise.
You will have {{ num_reflections }} rounds to iterate on the idea, but do not need to use them all.
""",
)

idea_refiner = Agent(
    name="Idea Refiner",
    instructions="""Round {{ current_round }}/{{ num_reflections }}.
In your thoughts, first carefully consider the quality, novelty, and feasibility of the idea you just created.
Include any other factors that you think are important in evaluating the idea.
Ensure the idea is clear and concise, and the JSON is the correct format.
Do not make things overly complicated.
In the next attempt, try and refine and improve your idea.
Stick to the spirit of the original idea unless there are glaring issues.

Respond in the same format as before:

{% if model not in reasoning_models %}
THOUGHT:
<THOUGHT>

NEW IDEA JSON:
{% endif %}
```json
<JSON>
```

If there is nothing to improve, simply repeat the previous JSON EXACTLY {% if model not in reasoning_models %}after the thought {% endif %}and include "I am done" at the end of the thoughts but before the JSON.
ONLY INCLUDE "I am done" IF YOU ARE MAKING NO MORE CHANGES.""",
)

idea = Swarm()
idea.add_node(0, idea_generator)
for i in range(5):
    idea.add_node(i + 1, idea_refiner)
    idea.add_edge(i, i + 1)



import json
import re
from datetime import datetime, timezone
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from openai.types.chat.chat_completion import ChatCompletion, Choice
from openai.types.chat.chat_completion_assistant_message_param import (
    ChatCompletionAssistantMessageParam,
)
from openai.types.chat.chat_completion_chunk import (
    ChatCompletionChunk,
    ChoiceDelta,
    ChoiceDeltaToolCall,
    ChoiceDeltaToolCallFunction,
)
from openai.types.chat.chat_completion_chunk import Choice as ChunkChoice
from openai.types.chat.chat_completion_message import ChatCompletionMessage


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(
        "--deepeval", action="store_true", default=False, help="Run deepeval tests"
    )
    parser.addoption(
        "--openai",
        action="store_true",
        default=False,
        help="Run tests with actual OpenAI API calls instead of mocks",
    )
    parser.addoption(
        "--model",
        action="store",
        default="deepseek-reasoner",
        help="Model to use for tests",
    )


@pytest.fixture
def skip_deepeval(pytestconfig: pytest.Config):
    return not pytestconfig.getoption("deepeval")


@pytest.fixture
def is_mocking(pytestconfig: pytest.Config):
    return not pytestconfig.getoption("openai")


@pytest.fixture
def anyio_backend():
    return "asyncio"


def create_mock_streaming_response(
    message: ChatCompletionAssistantMessageParam,
    model: str,
):
    content = message.get("content")
    match content:
        case str():
            tokens = re.split(r"(\s+)", content)
        case None:
            tokens = []
        case _:
            tokens = [part["text"] for part in content if part["type"] == "text"]

    tool_calls = list(message.get("tool_calls", []))

    async def _generator():
        for i, token in enumerate(tokens):
            yield ChatCompletionChunk(
                id="mock_cc_id",
                created=int(datetime.now().timestamp()),
                model=model,
                object="chat.completion.chunk",
                choices=[
                    ChunkChoice(
                        delta=ChoiceDelta(
                            content=token,
                            role="assistant" if i == 0 else None,
                        ),
                        finish_reason="stop"
                        if i == len(tokens) - 1 and not tool_calls
                        else None,
                        index=0,
                    )
                ],
            )
        for i, call in enumerate(tool_calls):
            yield ChatCompletionChunk(
                id="mock_cc_id",
                created=int(datetime.now().timestamp()),
                model=model,
                object="chat.completion.chunk",
                choices=[
                    ChunkChoice(
                        index=0,
                        delta=ChoiceDelta(
                            role="assistant" if i == 0 else None,
                            tool_calls=[
                                ChoiceDeltaToolCall(
                                    index=i,
                                    id=f"mock_tc_id_{i}",
                                    type="function",
                                    function=ChoiceDeltaToolCallFunction(
                                        name=call["function"]["name"],
                                        arguments=call["function"]["arguments"],
                                    ),
                                )
                            ],
                        ),
                    )
                ],
            )

    return _generator()


def create_mock_response(message, model):
    message = ChatCompletionMessage.model_validate(message)

    return ChatCompletion(
        id="mock_cc_id",
        created=int(datetime.now().timestamp()),
        model=model,
        object="chat.completion",
        choices=[
            Choice(
                message=message,
                finish_reason="tool_calls" if message.tool_calls else "stop",
                index=0,
            )
        ],
    )


class MockAsyncOpenAI:
    def __init__(self):
        self.chat = MagicMock()
        self.chat.completions = MagicMock()
        self.chat.completions.create = AsyncMock()

    def set_response(self, response: ChatCompletion):
        """
        Set the mock to return a specific response.
        :param response: A ChatCompletion response to return.
        """
        self.chat.completions.create.return_value = response

    def set_sequential_responses(self, responses: list[ChatCompletion]):
        """
        Set the mock to return different responses sequentially.
        :param responses: A list of ChatCompletion responses to return in order.
        """
        self.chat.completions.create.side_effect = responses


@pytest.fixture
def mock_openai():
    return MockAsyncOpenAI()


@pytest.fixture(autouse=True)
def client(
    mock_openai: MockAsyncOpenAI,
    is_mocking: bool,
    request: pytest.FixtureRequest,
    model: str,
):
    if is_mocking:
        match request.node.name:
            case "test_mcp_tool_call":
                messages = json.loads(
                    (
                        Path(__file__).parent / "threads" / "mcp_tool_call.json"
                    ).read_text()
                )
                mock_openai.set_sequential_responses(
                    [
                        create_mock_response(message, model)
                        for message in messages
                        if message["role"] == "assistant"
                    ]
                    + [
                        create_mock_response(
                            {
                                "content": datetime.now(timezone.utc).strftime(
                                    "%H:%M:%S"
                                ),
                                "role": "assistant",
                            },
                            model,
                        )
                    ]
                )
            case name if name.startswith("test_"):
                n = name.replace("test_", "")
                if n.endswith("_streaming"):
                    n = n.replace("_streaming", "")
                if (
                    json_file := Path(__file__).parent / "threads" / f"{n}.json"
                ).exists():
                    messages = json.loads(json_file.read_text())
                    if name.endswith("_streaming"):
                        mock_openai.set_sequential_responses(
                            [
                                create_mock_streaming_response(message, model)
                                for message in messages
                            ]
                        )
                    else:
                        mock_openai.set_sequential_responses(
                            [
                                create_mock_response(message, model)
                                for message in messages
                                if message["role"] == "assistant"
                            ]
                        )
    with patch("swarmx.DEFAULT_CLIENT", mock_openai):
        yield


@pytest.fixture
def model(pytestconfig: pytest.Config):
    return pytestconfig.getoption("model")



import datetime
import operator
import sys

import pytest
from mcp.client.stdio import StdioServerParameters

from swarmx import Agent, Swarm

pytestmark = pytest.mark.anyio


async def test_swarm_creation():
    """Test creating a basic swarm with a single agent node."""
    agent = Agent(name="test_agent", instructions="You are a helpful assistant.")

    swarm = Swarm()
    swarm.add_node(0, type="agent", agent=agent)

    assert 0 in swarm.nodes
    assert swarm.root == 0
    assert isinstance(swarm.nodes[0]["agent"], Agent)
    assert swarm.nodes[0]["agent"].name == "test_agent"


async def test_swarm_with_multiple_agents():
    """Test creating a swarm with multiple agents in a directed graph."""
    agent1 = Agent(name="agent1")
    agent2 = Agent(name="agent2")
    agent3 = Agent(name="agent3")

    swarm = Swarm()
    swarm.add_node(0, type="agent", agent=agent1)
    swarm.add_node(1, type="agent", agent=agent2)
    swarm.add_node(2, type="agent", agent=agent3)

    swarm.add_edge(0, 1)
    swarm.add_edge(0, 2)

    assert list(swarm.successors(0)) == [1, 2]
    assert swarm.root == 0
    assert len(swarm.nodes) == 3
    assert len(swarm.edges) == 2


async def test_swarm_serialization():
    """Test serializing and deserializing a swarm."""
    agent1 = Agent(name="agent1")
    agent2 = Agent(name="agent2")

    # Create a swarm
    swarm = Swarm()
    swarm.add_node(0, type="agent", agent=agent1)
    swarm.add_node(1, type="agent", agent=agent2)
    swarm.add_edge(0, 1, priority=1)

    # Serialize
    serialized = swarm.model_dump(mode="json")

    # Check serialization result
    assert "nodes" in serialized
    assert "edges" in serialized
    assert "graph" in serialized
    assert len(serialized["nodes"]) == 2
    assert len(serialized["edges"]) == 1
    assert serialized["edges"][0]["priority"] == 1

    # Deserialize
    deserialized = Swarm.model_validate(serialized)

    # Check graph structure is preserved
    assert list(deserialized.nodes) == [0, 1]
    assert list(deserialized.edges) == [(0, 1)]
    assert deserialized.edges[0, 1]["priority"] == 1
    assert deserialized.nodes[0]["agent"].name == "agent1"
    assert deserialized.nodes[1]["agent"].name == "agent2"


async def test_swarm_with_mcp_servers():
    """Test swarm with MCP servers configuration."""
    agent = Agent(name="agent")

    # Create a swarm with MCP server configuration
    swarm = Swarm(
        mcpServers={
            "time": StdioServerParameters(
                command=sys.executable,
                args=["-m", "mcp_server_time", "--local-timezone", "UTC"],
            )
        }
    )
    swarm.add_node(0, type="agent", agent=agent)

    # Check server config is present
    assert "time" in swarm.mcpServers
    assert swarm.mcpServers["time"].command == sys.executable
    assert swarm.mcpServers["time"].args == [
        "-m",
        "mcp_server_time",
        "--local-timezone",
        "UTC",
    ]

    # Test serialization with MCP servers
    serialized = swarm.model_dump(mode="json")
    assert "mcpServers" in serialized
    assert "time" in serialized["mcpServers"]


async def test_swarm_root_detection():
    """Test root node detection in a swarm."""
    agent1 = Agent(name="agent1")
    agent2 = Agent(name="agent2")
    agent3 = Agent(name="agent3")

    # Create a graph with one root
    swarm = Swarm()
    swarm.add_node(0, type="agent", agent=agent1)
    swarm.add_node(1, type="agent", agent=agent2)
    swarm.add_node(2, type="agent", agent=agent3)
    swarm.add_edge(0, 1)
    swarm.add_edge(1, 2)

    assert swarm.root == 0

    # Test error when multiple roots
    swarm2 = Swarm()
    swarm2.add_node(0, type="agent", agent=agent1)
    swarm2.add_node(1, type="agent", agent=agent2)

    with pytest.raises(ValueError, match="Swarm must have exactly one root node"):
        _ = swarm2.root


async def test_swarm_run_with_tool_execution():
    """Test swarm run with tool execution."""

    def calculator(a: int, b: int, operation: str) -> str:
        """Calculate the result of the operation."""
        return str(
            {
                "add": operator.add,
                "subtract": operator.sub,
                "multiply": operator.mul,
                "divide": operator.truediv,
            }[operation](a, b)
        )

    agent = Agent(name="agent", tools=[calculator])
    swarm = Swarm()
    swarm.add_node(0, type="agent", agent=agent)

    # Execute the run method
    messages = [
        {
            "role": "user",
            "content": "Calculate 5+3, just answer with the result without any other text",
        }
    ]
    new_messages = await swarm.run(
        messages=messages, context_variables={"initial": "context"}, stream=False
    )

    # Verify the result
    assert len(new_messages) == 3  # Assistant message + tool + assistant
    assert new_messages[0]["role"] == "assistant"
    assert new_messages[1]["role"] == "tool"
    assert new_messages[1]["content"] == "8"
    assert new_messages[2]["role"] == "assistant"
    assert new_messages[2]["content"] == "8"


async def test_mcp_tool_call():
    """Test swarm with MCP server integration."""
    agent = Agent(name="agent")

    # Create a swarm with MCP server
    swarm = Swarm(
        mcpServers={
            "time": StdioServerParameters(
                command=sys.executable,
                args=["-m", "mcp_server_time", "--local-timezone", "UTC"],
            )
        }
    )
    swarm.add_node(0, type="agent", agent=agent)

    messages = await swarm.run(
        messages=[
            {
                "role": "user",
                "content": "What's the time now? Answer me in 'HH:MM:SS' format",
            }
        ]
    )
    t = datetime.datetime.strptime(messages[-1]["content"], "%H:%M:%S")
    n = datetime.datetime.now(datetime.UTC)
    # make t as UTC
    t = t.astimezone(datetime.UTC)
    assert t - n < datetime.timedelta(seconds=2)


async def test_swarm_of_swarms():
    """Test swarm that contains another swarm."""
    # Create inner swarm
    inner_agent = Agent(name="inner_agent")
    inner_swarm = Swarm()
    inner_swarm.add_node(0, type="agent", agent=inner_agent)

    # Create outer swarm with inner swarm as a node
    agent = Agent(name="agent")
    outer_swarm = Swarm()
    outer_swarm.add_node(0, type="agent", agent=agent)
    outer_swarm.add_node(1, type="swarm", swarm=inner_swarm)
    outer_swarm.add_edge(0, 1)

    # Verify the structure
    assert "agent" in outer_swarm.nodes[0]
    assert "swarm" in outer_swarm.nodes[1]
    assert outer_swarm.nodes[1]["swarm"] == inner_swarm

    # Test serialization of nested swarm
    serialized = outer_swarm.model_dump(mode="json")
    assert len(serialized["nodes"]) == 2
    assert serialized["nodes"][1]["type"] == "swarm"
    assert "nodes" in serialized["nodes"][1]["swarm"]

    # Test deserialization of nested swarm
    deserialized = Swarm.model_validate(serialized)
    assert "swarm" in deserialized.nodes[1]
    assert isinstance(deserialized.nodes[1]["swarm"], Swarm)
    assert deserialized.nodes[1]["swarm"].nodes[0]["agent"].name == "inner_agent"


async def test_swarm_run_with_max_turns():
    """Test swarm run with max_turns limit."""
    agent = Agent(name="agent")
    swarm = Swarm()
    swarm.add_node(0, type="agent", agent=agent)

    # Run with max_turns=1
    messages = await swarm.run(
        messages=[{"role": "user", "content": "Test max turns"}], max_turns=1
    )

    # Verify only one turn was executed
    assert len(messages) == 1
    assert messages[0]["content"] == "First response"


async def test_swarm_validation_errors():
    """Test error handling in Swarm validation."""
    swarm = Swarm()

    # Test adding edge with non-existent nodes
    with pytest.raises(ValueError, match="Node 99 not found"):
        swarm.add_edge(99, 100)

    # Add a node first
    agent = Agent(name="agent")
    swarm.add_node(0, type="agent", agent=agent)

    # Test adding edge with one non-existent node
    with pytest.raises(ValueError, match="Node 99 not found"):
        swarm.add_edge(0, 99)

    # Test deserializing invalid data
    with pytest.raises(Exception):
        Swarm.model_validate({"nodes": [{"invalid": "data"}]})


async def test_handoff():
    client = Swarm()
    english_agent = Agent(
        name="English Agent",
        instructions="You only speak English.",
        functions=["tests.functions.transfer_to_spanish_agent"],
    )
    message_input = "Hola. ¿Como estás?"
    client.add_node(0, type="agent", agent=english_agent)
    messages = await client.run(
        messages=[{"role": "user", "content": message_input}],
    )
    assert messages[-1].get("name").startswith("Spanish Agent")


@pytest.fixture
def simple_swarm():
    swarm = Swarm()
    agent1 = Agent(name="Agent1", instructions="You are Agent 1")
    agent2 = Agent(name="Agent2", instructions="You are Agent 2")

    swarm.add_node(0, type="agent", agent=agent1)
    swarm.add_node(1, type="agent", agent=agent2)
    swarm.add_edge(0, 1)

    return swarm


def test_swarm_root(simple_swarm):
    assert simple_swarm.root == 0


def test_next_node(simple_swarm):
    assert simple_swarm._next_node == 2


def test_swarm_add_edge_validation():
    swarm = Swarm()
    agent1 = Agent(name="Agent1")
    swarm.add_node(0, type="agent", agent=agent1)

    # Test with non-existent source node
    with pytest.raises(ValueError, match="Node 999 not found"):
        swarm.add_edge(999, 0)

    # Test with non-existent target node
    with pytest.raises(ValueError, match="Node 999 not found"):
        swarm.add_edge(0, 999)


async def test_swarm_streaming():
    """Test swarm streaming functionality."""

    # Create a simple calculator agent
    def calculator(a: int, b: int, operation: str) -> str:
        """Calculate the result of the operation."""
        return str(
            {
                "add": operator.add,
                "subtract": operator.sub,
                "multiply": operator.mul,
                "divide": operator.truediv,
            }[operation](a, b)
        )

    agent = Agent(name="calculator", tools=[calculator])
    swarm = Swarm()
    swarm.add_node(0, type="agent", agent=agent)

    messages = [
        {
            "role": "user",
            "content": "Calculate 5+3, then multiply the result by 2",
        }
    ]

    chunks = []
    async for chunk in await swarm.run(messages=messages, stream=True):
        chunks.append(chunk)

    # Verify we got multiple chunks
    assert len(chunks) > 1

    # Verify the chunks contain both completion chunks and tool results
    from openai.types.chat import ChatCompletionChunk

    completion_chunks = [c for c in chunks if isinstance(c, ChatCompletionChunk)]
    assert len(completion_chunks) > 0

    # Verify the final result through merging chunks
    from swarmx import merge_chunks

    merged_messages = merge_chunks(completion_chunks)
    assert len(merged_messages) > 0

    # The final message should contain 16 (result of (5+3)*2)
    assert "16" in merged_messages[-1]["content"]


async def test_simple_chain():
    """Test a simple chain of agents: A -> B -> C"""
    swarm = Swarm()

    # Create chain A -> B -> C
    swarm.add_node(0, type="agent", agent=Agent(name="A"))
    swarm.add_node(1, type="agent", agent=Agent(name="B"))
    swarm.add_node(2, type="agent", agent=Agent(name="C"))
    swarm.add_edge(0, 1)
    swarm.add_edge(1, 2)

    messages = await swarm.run(messages=[{"role": "user", "content": "Test"}])

    assert swarm.nodes[2]["agent"].name == "C"
    assert messages[-1]["content"] == "C"


async def test_branch_execution():
    """Test branching execution: A -> (B, C)"""
    swarm = Swarm()

    # Create branches A -> B and A -> C
    swarm.add_node(0, type="agent", agent=Agent(name="A"))
    swarm.add_node(1, type="agent", agent=Agent(name="B"))
    swarm.add_node(2, type="agent", agent=Agent(name="C"))
    swarm.add_edge(0, 1)
    swarm.add_edge(0, 2)

    messages = await swarm.run(messages=[{"role": "user", "content": "Test"}])
    assert len(messages) == 3
    assert messages[0]["content"] == "A"


async def test_complex_graph():
    """Test complex graph with multiple branches and joins"""
    swarm = Swarm()

    # Create a more complex graph:
    # A -> B -> D
    # A -> C -> D
    # D -> E
    swarm.add_node(0, type="agent", agent=Agent(name="A"))
    swarm.add_node(1, type="agent", agent=Agent(name="B"))
    swarm.add_node(2, type="agent", agent=Agent(name="C"))
    swarm.add_node(3, type="agent", agent=Agent(name="D"))
    swarm.add_node(4, type="agent", agent=Agent(name="E"))

    swarm.add_edge(0, 1)  # A -> B
    swarm.add_edge(0, 2)  # A -> C
    swarm.add_edge(1, 3)  # B -> D
    swarm.add_edge(2, 3)  # C -> D
    swarm.add_edge(3, 4)  # D -> E

    messages = [{"role": "user", "content": "Test"}]
    await swarm.run(messages=messages)


async def test_conditional_edges():
    """Test conditional edges in a swarm."""
    swarm = Swarm()

    # Create agents
    agent1 = Agent(name="Agent1")
    agent2 = Agent(name="Agent2")
    agent3 = Agent(name="Agent3")

    # Add nodes
    swarm.add_node(0, type="agent", agent=agent1)
    swarm.add_node(1, type="agent", agent=agent2)
    swarm.add_node(2, type="agent", agent=agent3)

    # Add conditional edges
    def condition_true(context):
        return True

    def condition_false(context):
        return False

    def condition_based_on_context(context):
        return context.get("should_go_to_agent2", False)

    # Add edges with conditions
    swarm.add_edge(0, 1, condition=condition_true)  # Always goes to agent2
    swarm.add_edge(0, 2, condition=condition_false)  # Never goes to agent3

    # Test that only agent2 is reached
    messages = await swarm.run(messages=[{"role": "user", "content": "Test"}])
    assert len(messages) == 2  # Agent1 response + Agent2 response
    assert messages[-1]["name"].startswith("Agent2")

    # Test with context-based condition
    swarm = Swarm()
    swarm.add_node(0, type="agent", agent=agent1)
    swarm.add_node(1, type="agent", agent=agent2)
    swarm.add_node(2, type="agent", agent=agent3)

    swarm.add_edge(0, 1, condition=condition_based_on_context)
    swarm.add_edge(0, 2, condition=lambda ctx: not condition_based_on_context(ctx))

    # Test with context that should go to agent2
    messages = await swarm.run(
        messages=[{"role": "user", "content": "Test"}],
        context_variables={"should_go_to_agent2": True},
    )
    assert len(messages) == 2  # Agent1 response + Agent2 response
    assert messages[-1]["name"].startswith("Agent2")

    # Test with context that should go to agent3
    messages = await swarm.run(
        messages=[{"role": "user", "content": "Test"}],
        context_variables={"should_go_to_agent2": False},
    )
    assert len(messages) == 2  # Agent1 response + Agent3 response
    assert messages[-1]["name"].startswith("Agent3")



from swarmx import Agent


def transfer_to_spanish_agent() -> Agent:
    return Agent(
        name="Spanish Agent",
        instructions="You only speak Spanish.",
    )


def print_account_details(context_variables: dict):
    """Simple function to print account details."""
    user_id = context_variables.get("user_id", None)
    name = context_variables.get("name", None)
    return f"Account Details: {name} {user_id}"



from collections import defaultdict
from typing import Annotated, Any

import mcp.types
import pytest
from openai.types.chat.chat_completion_chunk import (
    ChatCompletionChunk,
    ChoiceDelta,
    ChoiceDeltaToolCall,
    ChoiceDeltaToolCallFunction,
)
from openai.types.chat.chat_completion_chunk import (
    Choice as ChunkChoice,
)

from swarmx import (
    Agent,
    Result,
    _image_content_to_url,
    _mcp_call_tool_result_to_content,
    _resource_to_file,
    check_instructions,
    content_part_to_str,
    function_to_json,
    merge_chunk,
    merge_chunks,
    messages_to_chunks,
    validate_tool,
)

from .functions import print_account_details

pytestmark = pytest.mark.anyio


def sample(a: int, b: str) -> str:
    """Sample function"""
    return f"{a}{b}"


async def async_func(c: float) -> Agent:
    return Agent()


def no_params() -> dict:
    return {}


class TestFunctionToJson:
    def test_function_to_openai_tool(self):
        assert function_to_json(print_account_details) == {
            "function": {
                "description": "Simple function to print account details.",
                "name": "print_account_details",
                "parameters": {"properties": {}, "type": "object"},
            },
            "type": "function",
        }

    def test_basic_function_conversion(self):
        tool = function_to_json(sample)
        assert tool["type"] == "function"
        assert tool["function"]["name"] == "sample"
        assert tool["function"]["description"] == "Sample function"
        assert tool["function"]["parameters"] == {
            "type": "object",
            "properties": {"a": {"type": "integer"}, "b": {"type": "string"}},
            "required": ["a", "b"],
        }

    def test_context_variables_exclusion(self):
        def with_context(a: int, context_variables: dict) -> str:
            return ""

        tool = function_to_json(with_context)
        assert "context_variables" not in tool["function"]["parameters"]["properties"]

    def test_async_function_handling(self):
        tool = function_to_json(async_func)
        assert tool["function"]["name"] == "async_func"
        assert tool["function"]["parameters"] == {
            "type": "object",
            "properties": {"c": {"type": "number"}},
            "required": ["c"],
        }

    def test_function_with_complex_types(self):
        def complex_types(
            d: Annotated[str, "metadata"], e: list[dict[str, int]]
        ) -> Result:
            """Complex types"""
            return Result(content=[])

        tool = function_to_json(complex_types)
        assert tool["function"]["parameters"] == {
            "type": "object",
            "properties": {
                "d": {"type": "string"},
                "e": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "additionalProperties": {"type": "integer"},
                    },
                },
            },
            "required": ["d", "e"],
        }

    def test_function_with_no_parameters(self):
        tool = function_to_json(no_params)
        assert tool["function"]["parameters"] == {"type": "object", "properties": {}}


def test_merge_content_string():
    message = {"role": "assistant", "content": "Hello"}
    delta = ChoiceDelta(content=" world")
    merge_chunk(message, delta)
    assert message["content"] == "Hello world"


def test_merge_content_list():
    message = {"role": "assistant", "content": [{"type": "text", "text": "Hi"}]}
    delta = ChoiceDelta(content=" there")
    merge_chunk(message, delta)
    assert message["content"] == [
        {"type": "text", "text": "Hi"},
        {"type": "text", "text": " there"},
    ]


def test_merge_refusal():
    message = {"role": "assistant", "refusal": "No"}
    delta = ChoiceDelta(refusal=" way")
    merge_chunk(message, delta)
    assert message["refusal"] == "No way"


def test_new_tool_call():
    message = {
        "role": "assistant",
        "tool_calls": defaultdict(
            lambda: {
                "id": "",
                "type": "function",
                "function": {"arguments": "", "name": ""},
            }
        ),
    }
    delta = ChoiceDelta(
        tool_calls=[
            ChoiceDeltaToolCall(
                index=0,
                id="call_1",
                function=ChoiceDeltaToolCallFunction(arguments="arg", name="func"),
            )
        ]
    )
    merge_chunk(message, delta)
    assert message["tool_calls"][0] == {
        "id": "call_1",
        "function": {"arguments": "arg", "name": "func"},
        "type": "function",
    }


def test_update_existing_tool_call():
    message = {
        "role": "assistant",
        "tool_calls": [
            {
                "id": "call_1",
                "function": {"arguments": "arg1", "name": ""},
                "type": "function",
            }
        ],
    }
    delta = ChoiceDelta(
        tool_calls=[
            ChoiceDeltaToolCall(
                index=0,
                function=ChoiceDeltaToolCallFunction(arguments="arg2", name="func2"),
            )
        ]
    )
    merge_chunk(message, delta)
    assert message["tool_calls"][0]["function"]["arguments"] == "arg1arg2"
    assert message["tool_calls"][0]["function"]["name"] == "func2"


def test_multiple_tool_calls():
    message = {
        "role": "assistant",
        "tool_calls": defaultdict(
            lambda: {
                "id": "",
                "type": "function",
                "function": {"arguments": "", "name": ""},
            }
        ),
    }
    delta = ChoiceDelta(
        tool_calls=[
            ChoiceDeltaToolCall(
                index=0,
                id="call_1",
                function=ChoiceDeltaToolCallFunction(arguments="arg1", name="func1"),
            ),
            ChoiceDeltaToolCall(
                index=1,
                id="call_2",
                function=ChoiceDeltaToolCallFunction(arguments="arg2", name="func2"),
            ),
        ]
    )
    merge_chunk(message, delta)
    assert len(message["tool_calls"]) == 2
    assert message["tool_calls"][0]["id"] == "call_1"
    assert message["tool_calls"][1]["id"] == "call_2"


def test_merge_chunk_with_content():
    """Test merging chunk with content into a message."""
    message = {
        "role": "assistant",
        "content": "Hello, ",
        "reasoning_content": "Thinking... ",
    }

    delta = ChoiceDelta(content=" world!", reasoning_content=" Almost done!")

    merge_chunk(message, delta)

    assert message["content"] == "Hello,  world!"
    assert message["reasoning_content"] == "Thinking...  Almost done!"


def test_merge_chunk_with_content_parts():
    """Test merging chunk with content parts into a message that already has content parts."""
    message = {
        "role": "assistant",
        "content": [{"type": "text", "text": "Hello, "}],
        "reasoning_content": [{"type": "text", "text": "Thinking... "}],
    }

    delta = ChoiceDelta(content=" world!", reasoning_content=" Almost done!")

    merge_chunk(message, delta)

    assert len(message["content"]) == 2
    assert message["content"][0]["text"] == "Hello, "
    assert message["content"][1]["text"] == " world!"

    assert len(message["reasoning_content"]) == 2
    assert message["reasoning_content"][0]["text"] == "Thinking... "
    assert message["reasoning_content"][1]["text"] == " Almost done!"


def test_merge_chunk_with_refusal():
    """Test merging chunk with refusal into a message."""
    message = {"role": "assistant"}

    delta = ChoiceDelta(refusal="I cannot provide information about that topic.")

    merge_chunk(message, delta)

    assert message["refusal"] == "I cannot provide information about that topic."

    # Test appending to existing refusal
    delta2 = ChoiceDelta(refusal=" Please ask something else.")
    merge_chunk(message, delta2)

    assert (
        message["refusal"]
        == "I cannot provide information about that topic. Please ask something else."
    )


def test_merge_chunk_with_tool_calls():
    """Test merging chunk with tool calls into a message."""
    message = {
        "role": "assistant",
        "tool_calls": defaultdict(
            lambda: {
                "id": "",
                "type": "function",
                "function": {"arguments": "", "name": ""},
            }
        ),
    }

    # First tool call chunk
    delta = ChoiceDelta(
        tool_calls=[
            ChoiceDeltaToolCall(
                index=0,
                id="call_123",
                function=ChoiceDeltaToolCallFunction(
                    name="search",
                    arguments='{"query": "',
                ),
                type="function",
            )
        ]
    )

    merge_chunk(message, delta)

    assert len(message["tool_calls"]) == 1
    assert message["tool_calls"][0]["id"] == "call_123"
    assert message["tool_calls"][0]["function"]["name"] == "search"
    assert message["tool_calls"][0]["function"]["arguments"] == '{"query": "'

    # Continuation of the first tool call
    delta2 = ChoiceDelta(
        tool_calls=[
            ChoiceDeltaToolCall(
                index=0,
                id="call_123",
                function=ChoiceDeltaToolCallFunction(arguments='weather forecast"}'),
                type="function",
            )
        ]
    )

    merge_chunk(message, delta2)

    assert len(message["tool_calls"]) == 1
    assert message["tool_calls"][0]["id"] == "call_123"
    assert message["tool_calls"][0]["function"]["name"] == "search"
    assert (
        message["tool_calls"][0]["function"]["arguments"]
        == '{"query": "weather forecast"}'
    )

    # Second tool call
    delta3 = ChoiceDelta(
        tool_calls=[
            ChoiceDeltaToolCall(
                index=1,
                id="call_456",
                function=ChoiceDeltaToolCallFunction(name="get_time", arguments="{}"),
                type="function",
            )
        ]
    )

    merge_chunk(message, delta3)

    assert len(message["tool_calls"]) == 2
    assert message["tool_calls"][0]["id"] == "call_123"
    assert message["tool_calls"][1]["id"] == "call_456"
    assert message["tool_calls"][1]["function"]["name"] == "get_time"
    assert message["tool_calls"][1]["function"]["arguments"] == "{}"


def test_merge_chunk_multiple_updates():
    """Test merging multiple types of updates into a message."""
    message = {
        "role": "assistant",
        "content": "Initial content. ",
        "reasoning_content": "Initial reasoning. ",
    }

    delta = ChoiceDelta(
        content="Additional content.",
        reasoning_content="Additional reasoning.",
        refusal="This is a refusal message.",
        tool_calls=[
            ChoiceDeltaToolCall(
                index=0,
                id="tool_call_1",
                function=ChoiceDeltaToolCallFunction(
                    name="test_tool", arguments='{"param": "value"}'
                ),
                type="function",
            )
        ],
    )

    merge_chunk(message, delta)

    assert message["content"] == "Initial content. Additional content."
    assert message["reasoning_content"] == "Initial reasoning. Additional reasoning."
    assert message["refusal"] == "This is a refusal message."
    assert len(message["tool_calls"]) == 1
    assert message["tool_calls"][0]["id"] == "tool_call_1"
    assert message["tool_calls"][0]["function"]["name"] == "test_tool"
    assert message["tool_calls"][0]["function"]["arguments"] == '{"param": "value"}'


def test_merge_chunks():
    """Test merging multiple chunks into a list of messages."""

    chunks = [
        ChatCompletionChunk(
            id="chunk_1",
            choices=[
                ChunkChoice(
                    index=0,
                    delta=ChoiceDelta(
                        role="assistant",
                        content="Hello, ",
                        reasoning_content="Thinking about greeting the user... ",
                    ),
                    finish_reason=None,
                )
            ],
            created=1716000000,
            model="test-model",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chunk_1",
            choices=[
                ChunkChoice(
                    index=0,
                    delta=ChoiceDelta(
                        content="world!",
                        reasoning_content="Decided to use a simple greeting.",
                    ),
                    finish_reason=None,
                )
            ],
            created=1716000001,
            model="test-model",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chunk_1",
            choices=[
                ChunkChoice(
                    index=0,
                    delta=ChoiceDelta(
                        tool_calls=[
                            ChoiceDeltaToolCall(
                                index=0,
                                id="call_abc",
                                function=ChoiceDeltaToolCallFunction(
                                    name="get_time", arguments="{}"
                                ),
                                type="function",
                            )
                        ]
                    ),
                    finish_reason=None,
                )
            ],
            created=1716000002,
            model="test-model",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chunk_1",
            choices=[
                ChunkChoice(
                    index=0,
                    delta=ChoiceDelta(),
                    finish_reason="stop",
                )
            ],
            created=1716000003,
            model="test-model",
            object="chat.completion.chunk",
        ),
    ]

    # Run the merge_chunks function with our async generator
    messages = merge_chunks(chunks)

    # Verify results
    assert len(messages) == 1
    message = messages[0]
    assert message["role"] == "assistant"
    assert message["content"] == "Hello, world!"
    assert (
        message["reasoning_content"]
        == "Thinking about greeting the user... Decided to use a simple greeting."
    )
    assert len(message["tool_calls"]) == 1
    assert message["tool_calls"][0]["id"] == "call_abc"
    assert message["tool_calls"][0]["function"]["name"] == "get_time"
    assert message["tool_calls"][0]["function"]["arguments"] == "{}"


def test_merge_chunks_multiple_messages():
    """Test merging multiple chunks with different IDs into multiple messages."""

    chunks = [
        ChatCompletionChunk(
            id="msg_1",
            choices=[
                ChunkChoice(
                    index=0,
                    delta=ChoiceDelta(role="assistant", content="First message"),
                    finish_reason=None,
                )
            ],
            created=1716000000,
            model="test-model",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="msg_2",
            choices=[
                ChunkChoice(
                    index=0,
                    delta=ChoiceDelta(role="assistant", content="Second message"),
                    finish_reason=None,
                )
            ],
            created=1716000001,
            model="test-model",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="msg_1",
            choices=[
                ChunkChoice(
                    index=0,
                    delta=ChoiceDelta(),
                    finish_reason="stop",
                )
            ],
            created=1716000002,
            model="test-model",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="msg_2",
            choices=[
                ChunkChoice(
                    index=0,
                    delta=ChoiceDelta(),
                    finish_reason="stop",
                )
            ],
            created=1716000003,
            model="test-model",
            object="chat.completion.chunk",
        ),
    ]
    # Run the merge_chunks function
    messages = merge_chunks(chunks)

    # Sort messages by ID to ensure consistent order for testing
    messages.sort(key=lambda x: x.get("id", ""))

    # Verify results
    assert len(messages) == 2
    assert messages[0]["content"] == "First message"
    assert messages[1]["content"] == "Second message"


class TestCheckInstructions:
    def test_valid_string(self):
        instructions = "You are a helpful assistant"
        result = check_instructions(instructions)
        assert result == instructions

    def test_valid_callable(self):
        def valid_func(context_variables: dict) -> str:
            return "Hello"

        result = check_instructions(valid_func)
        assert result == valid_func

    def test_non_callable_non_string(self):
        with pytest.raises(ValueError) as excinfo:
            check_instructions(42)
        assert "a string or a callable" in str(excinfo.value)

    def test_generic_dict_annotation(self):
        def generic_anno(context_variables: dict[str, str]) -> str:
            return ""

        result = check_instructions(generic_anno)
        assert result == generic_anno

    def test_no_annotation_callable(self):
        def no_anno(context_variables):
            return ""

        result = check_instructions(no_anno)
        assert result == no_anno


class TestValidateTool:
    def test_valid_string_return(self):
        def valid_str() -> str:
            return "valid"

        assert validate_tool(valid_str)["function"]["name"] == "valid_str"

    def test_valid_agent_return(self):
        def valid_agent() -> Agent:
            return Agent()

        assert validate_tool(valid_agent)["function"]["name"] == "valid_agent"

    def test_valid_dict_return(self):
        def valid_dict() -> dict[str, Any]:
            return {"key": "value"}

        assert validate_tool(valid_dict)["function"]["name"] == "valid_dict"

    def test_valid_result_return(self):
        def valid_result() -> Result:
            return Result(content=[])

        assert validate_tool(valid_result)["function"]["name"] == "valid_result"

    def test_unannotated_return(self):
        def unannotated():
            return "no annotation"

        with pytest.warns(FutureWarning):
            checked = validate_tool(unannotated)
        assert checked["function"]["name"] == "unannotated"

    def test_invalid_return_type(self):
        def invalid_return() -> int:
            return 42

        with pytest.raises(TypeError) as excinfo:
            validate_tool(invalid_return)
        assert "must be str, Agent, dict[str, Any], or Result" in str(excinfo.value)

    def test_non_callable_input(self):
        with pytest.raises(TypeError) as excinfo:
            validate_tool(None)
        assert "must be str, Agent, dict[str, Any], or Result" in str(excinfo.value)

    def test_none_return_annotation(self):
        def none_return() -> None:
            return None

        with pytest.raises(TypeError) as excinfo:
            validate_tool(none_return)
        assert "must be str, Agent, dict[str, Any], or Result" in str(excinfo.value)


def test_image_content_to_url():
    """Test converting MCP image content to OpenAI image URL content part."""
    # Create a mock ImageContent
    image_content = mcp.types.ImageContent(
        type="image", mimeType="image/png", data="base64encodeddata"
    )

    # Convert to OpenAI format
    result = _image_content_to_url(image_content)

    # Verify result
    assert result["type"] == "image_url"
    assert result["image_url"]["url"] == "data:image/png;base64,base64encodeddata"


def test_resource_to_file_text():
    """Test converting MCP text resource to OpenAI text content part."""
    # Create a mock TextResourceContents
    text_content = mcp.types.TextResourceContents(
        uri="file:///some/path", mimeType="text/plain", text="Sample text"
    )

    # Create a mock EmbeddedResource with TextResourceContents
    resource = mcp.types.EmbeddedResource(type="resource", resource=text_content)

    # Convert to OpenAI format
    result = _resource_to_file(resource)

    # Verify result
    assert result["type"] == "text"
    assert result["text"] == '\n```text title="file:///some/path"\nSample text\n```\n'


def test_resource_to_file_blob():
    """Test converting MCP blob resource to OpenAI file content part."""
    # Create a mock BlobResourceContents
    blob_content = mcp.types.BlobResourceContents(
        uri="file:///some/path", mimeType="application/pdf", blob="base64encodedblob"
    )

    # Create a mock EmbeddedResource with BlobResourceContents
    resource = mcp.types.EmbeddedResource(type="resource", resource=blob_content)

    # Convert to OpenAI format
    result = _resource_to_file(resource)

    # Verify result
    assert result["type"] == "file"
    assert result["file"]["file_data"] == "base64encodedblob"


def test_content_part_to_str():
    """Test content_part_to_str function with different content part types."""
    # Test text part
    text_part = {"type": "text", "text": "hello"}
    assert content_part_to_str(text_part) == "hello"

    # Test refusal part
    refusal_part = {"type": "refusal", "refusal": "no"}
    assert content_part_to_str(refusal_part) == "no"

    # Test image_url part
    image_part = {
        "type": "image_url",
        "image_url": {"url": "http://example.com/image.png"},
    }
    assert content_part_to_str(image_part) == "![](http://example.com/image.png)"

    # Test unknown type
    unknown_part = {
        "type": "file",
        "file": {"file_data": "SGVsbG8sIFdvcmxkIQ==", "filename": "example.txt"},
    }
    assert (
        content_part_to_str(unknown_part)
        == '\n```text title="example.txt"\nHello, World!\n```\n'
    )

    binary_part = {
        "type": "file",
        "file": {
            "file_data": "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAAApgAAAKYB3X3/OAAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAANCSURBVEiJtZZPbBtFFMZ/M7ubXdtdb1xSFyeilBapySVU8h8OoFaooFSqiihIVIpQBKci6KEg9Q6H9kovIHoCIVQJJCKE1ENFjnAgcaSGC6rEnxBwA04Tx43t2FnvDAfjkNibxgHxnWb2e/u992bee7tCa00YFsffekFY+nUzFtjW0LrvjRXrCDIAaPLlW0nHL0SsZtVoaF98mLrx3pdhOqLtYPHChahZcYYO7KvPFxvRl5XPp1sN3adWiD1ZAqD6XYK1b/dvE5IWryTt2udLFedwc1+9kLp+vbbpoDh+6TklxBeAi9TL0taeWpdmZzQDry0AcO+jQ12RyohqqoYoo8RDwJrU+qXkjWtfi8Xxt58BdQuwQs9qC/afLwCw8tnQbqYAPsgxE1S6F3EAIXux2oQFKm0ihMsOF71dHYx+f3NND68ghCu1YIoePPQN1pGRABkJ6Bus96CutRZMydTl+TvuiRW1m3n0eDl0vRPcEysqdXn+jsQPsrHMquGeXEaY4Yk4wxWcY5V/9scqOMOVUFthatyTy8QyqwZ+kDURKoMWxNKr2EeqVKcTNOajqKoBgOE28U4tdQl5p5bwCw7BWquaZSzAPlwjlithJtp3pTImSqQRrb2Z8PHGigD4RZuNX6JYj6wj7O4TFLbCO/Mn/m8R+h6rYSUb3ekokRY6f/YukArN979jcW+V/S8g0eT/N3VN3kTqWbQ428m9/8k0P/1aIhF36PccEl6EhOcAUCrXKZXXWS3XKd2vc/TRBG9O5ELC17MmWubD2nKhUKZa26Ba2+D3P+4/MNCFwg59oWVeYhkzgN/JDR8deKBoD7Y+ljEjGZ0sosXVTvbc6RHirr2reNy1OXd6pJsQ+gqjk8VWFYmHrwBzW/n+uMPFiRwHB2I7ih8ciHFxIkd/3Omk5tCDV1t+2nNu5sxxpDFNx+huNhVT3/zMDz8usXC3ddaHBj1GHj/As08fwTS7Kt1HBTmyN29vdwAw+/wbwLVOJ3uAD1wi/dUH7Qei66PfyuRj4Ik9is+hglfbkbfR3cnZm7chlUWLdwmprtCohX4HUtlOcQjLYCu+fzGJH2QRKvP3UNz8bWk1qMxjGTOMThZ3kvgLI5AzFfo379UAAAAASUVORK5CYII=",
            "filename": "emoji.png",
        },
    }
    download_link = content_part_to_str(binary_part)
    assert download_link.startswith(
        "[emoji.png](data:application/octet-stream;base64,"
    ) and download_link.endswith(")")


def test_messages_to_chunks():
    """Test messages_to_chunks function with different message types."""
    # Test simple text message
    text_message = {"role": "assistant", "content": "hello"}
    chunks = list(messages_to_chunks([text_message]))
    assert len(chunks) == 1
    assert chunks[0].choices[0].delta.content == "hello"
    assert chunks[0].choices[0].finish_reason == "stop"

    # Test message with multiple content parts
    multi_part_message = {
        "role": "assistant",
        "content": [
            {"type": "text", "text": "hello"},
            {"type": "text", "text": " world"},
        ],
    }
    chunks = list(messages_to_chunks([multi_part_message]))
    assert len(chunks) == 2
    assert chunks[0].choices[0].delta.content == "hello"
    assert chunks[1].choices[0].delta.content == " world"
    assert chunks[1].choices[0].finish_reason == "stop"

    # Test message with refusal
    refusal_message = {"role": "assistant", "refusal": "no"}
    chunks = list(messages_to_chunks([refusal_message]))
    assert len(chunks) == 1
    assert chunks[0].choices[0].delta.refusal == "no"
    assert chunks[0].choices[0].finish_reason == "content_filter"

    # Test message with tool calls
    tool_message = {
        "role": "assistant",
        "tool_calls": [
            {
                "id": "1",
                "type": "function",
                "function": {"name": "test", "arguments": "{}"},
            }
        ],
    }
    chunks = list(messages_to_chunks([tool_message]))
    assert len(chunks) == 1
    assert chunks[0].choices[0].delta.tool_calls[0].function.name == "test"
    assert chunks[0].choices[0].finish_reason == "tool_calls"


def test_mcp_call_tool_result_to_content():
    """Test converting MCP CallToolResult to OpenAI content parts."""
    # Create a mock CallToolResult with various content types
    call_result = mcp.types.CallToolResult(
        content=[
            mcp.types.TextContent(type="text", text="Some text"),
            mcp.types.ImageContent(
                type="image", mimeType="image/jpeg", data="base64imagedata"
            ),
            mcp.types.EmbeddedResource(
                type="resource",
                resource=mcp.types.TextResourceContents(
                    uri="file:///some/path", mimeType="text/plain", text="Resource text"
                ),
            ),
        ]
    )

    result = _mcp_call_tool_result_to_content(call_result)

    # Verify results
    assert len(result) == 3
    assert result[0] == {"type": "text", "text": "Some text"}
    assert result[1]["type"] == "image_url"
    assert result[2]["type"] == "text"
